import weakref
import inspect

try:
    from weakref import WeakMethod

except ImportError:
    from .compat import WeakMethod


# ------------------------------------------------------------------------------
class Signal(object):

    # --------------------------------------------------------------------------
    def __init__(self):
        self._slots = list()

    # --------------------------------------------------------------------------
    def connect(self, slot):
        """
        Connects the given callable to the signal such that when the signal
        is emitted, any connecting slots will be called.

        Each connected slot will be called with any arguments that are passed
        through on the emit.

        :param slot: callable
        :return: True on successful connection
        """
        if not callable(slot):
            return False

        # -- Store the slot as a weak reference to ensure
        # -- this class does not go out of scope
        if inspect.ismethod(slot):
            self._slots.append(WeakMethod(slot))
        else:
            self._slots.append(weakref.ref(slot))

    # --------------------------------------------------------------------------
    def disconnect(self, slot):
        """
        Removes the given slot from the slot list meaning it will never be
        called when the signal is emitted.

        :param slot: callable to remove

        :return: True on successful removal
        """
        callables_to_remove = [
            possibility
            for possibility in self._slots
            if possibility() and possibility() == slot
        ]

        if not callables_to_remove:
            return False

        while callables_to_remove:
            self._slots.remove(callables_to_remove.pop())

        return True

    # --------------------------------------------------------------------------
    def emit(self, *args, **kwargs):
        """
        Triggers a call of all the connected slots passing through any
        arguments you give down to each slot.

        :param args: Args to be passed to each slot
        :param kwargs: Kwargs to be passed to each slot

        :return: None
        """
        for slot in self._slots:
            if slot():
                slot()(
                    *args,
                    **kwargs
                )



"""
This example demonstrates a scenario where you have a functional class which
carries out a process along with an accompanying interface (impletemented using
PySide2).

The Renamer class utilises recollection.Inference in order to automatically
store any property changes, allowing its state to be retained between sessions.

The choice for putting the recollection code in the Renamer rather than the
UI is a choice of implementation (putting it in the UI would be equally
valid). For this scenario the choice of putting it in the Renamer was to show
how you could easily get state persistent across different interfaces if you 
were to present the same renamer class in different ways based on the intended
user.
"""
from functools import partial

import os
import recollection


# -- We wrap the pyside import in a try except only
# -- because we have no direct dependency on this module
# -- outside of this example.
try:
    from PySide2 import QtWidgets

except ImportError:
    print('You must have PySide2 in order to run this example')


# ------------------------------------------------------------------------------
class FileRenamer(recollection.Inference):
    """
    This is a small renamer class. You can set the properties on the class 
    and call 'run' to invoke a renaming process. 
    
    Any settings you define will automatically be restored the next time
    you instance the object. 
    """
    # --------------------------------------------------------------------------
    def __init__(self):
        super(FileRenamer, self).__init__()

        # -- Input properties
        self.directory = ''
        self.suffix_filter = ''
        self.replace_this = ''
        self.replace_with = ''
        self.skip_readonly = True

        # -- Output properties
        self.time_taken = 0
        self.processed = list()
        self.failed = list()

        # -- This final block of code in the init is all that
        # -- is required to get persistent data serialisation
        # -- working. Because we choose to do this in the pure-code
        # -- class, every UI we make for it will automatically
        # -- benefit from persistent state functionality.
        self.memento.register(
            [
                'directory',
                'suffix_filter',
                'replace_this',
                'replace_with',
                'skip_readonly',
                'time_taken',
                'processed',
                'failed',
            ]
        )

        self.memento.register_serialiser(
            recollection.JsonAppSerialiser,
            'memento/examples/renamer',
            always_serialise=True,
        )
        self.memento.deserialise()

    # --------------------------------------------------------------------------
    def run(self):
        target_list = list()

        # -- Get a list of all the files we think we need
        # -- to operate one
        for root, _, filenames in os.walk(self.directory):
            for filename in filenames:

                # -- If we have a suffix test, validate that and skip
                # -- if it does not match
                if self.suffix_filter and self.suffix_filter not in filename:
                    continue

                # -- Check if this has a substring we need to replace, if
                # -- not we skip over it
                if self.replace_this not in filename:
                    continue

                target_list.append(
                    os.path.join(
                        root,
                        filename,
                    )
                )

        # -- Now we can operate on all the files and update the
        # -- log. Because we will be editing properties we have
        # -- asked recollection to watch, we ask it to defer its
        # -- store call until its complete.
        with self.memento.defer():
            # -- Clear the process list
            self.processed = list()
            self.failed = list()

            # -- Get a full path to the file
            for current_path in target_list:

                # -- There are a lot of things that can go wrong
                # -- during a rename, so wrap this in a try
                try:

                    # -- Define our new location
                    new_path = os.path.join(
                        os.path.dirname(current_path),
                        os.path.basename(current_path).replace(
                            self.replace_this,
                            self.replace_with,
                        ),
                    )

                    os.rename(
                        current_path,
                        new_path,
                    )

                    # -- Mark the fact that we processed this file
                    # -- successfully
                    self.processed.append(
                        '%s -> %s' % (
                            current_path,
                            new_path,
                        )
                    )

                except:
                    # -- Something went wrong, so log the fact that we
                    # -- did not rename this file
                    self.failed.append(
                        '%s could not be renamed' % current_path,
                    )


# ------------------------------------------------------------------------------
class FileRenamerUi(QtWidgets.QWidget):
    """
    This is the Ui element of the renamer. The Ui instances the Renamer object
    and then pushes any input from the user onto the renamer itself. 
    
    The Ui element does not store any information itself, instead it has 
    functionality to set its ui state based on the state of the renamer
    properties along with a mechanism to push any ui-triggered values into
    the renamer.
    
    By taking this approach the recollection of data is all handled in the
    Renamer class directly, and therefore exposes the ability to have multiple
    different interfaces aimed at different users without having to reimplement
    the storage mechanisms in each one.
    """
    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(FileRenamerUi, self).__init__(parent=parent)

        # -- Instance the renamer object
        self.renamer = FileRenamer()

        # -- Define our maine layout
        self.setLayout(QtWidgets.QVBoxLayout())

        # -- Create the directory layout
        self.directory_field = QtWidgets.QLineEdit()
        self.browser_button = QtWidgets.QPushButton('...')

        # -- Add the directory-centric widgets into this layout
        directory_layout = QtWidgets.QHBoxLayout()
        directory_layout.addWidget(self.directory_field)
        directory_layout.addWidget(self.browser_button)
        self.layout().addLayout(directory_layout)

        # -- Add the options
        options_group = QtWidgets.QGroupBox('Options')
        options_layout = QtWidgets.QVBoxLayout()
        options_group.setLayout(options_layout)
        self.layout().addWidget(options_group)

        # -- Utilise a private function which will handle the creation
        # -- of the layout and widget (and bind them together)
        layout_a, self.suffix_filter = self._labelled_lineedit('Suffix Filter')
        layout_b, self.replace_this = self._labelled_lineedit('Replace This')
        layout_c, self.with_this = self._labelled_lineedit('With This')

        options_layout.addLayout(layout_a)
        options_layout.addLayout(layout_b)
        options_layout.addLayout(layout_c)

        # -- Add our log output - which will show the results of the
        # -- process runner
        self.log_output = QtWidgets.QListWidget()
        self.layout().addWidget(self.log_output)

        # -- Add the button which will ultimately trigger the
        # -- processing
        self.run_button = QtWidgets.QPushButton('Process')
        self.layout().addWidget(self.run_button)

        # -- Update the Ui based on the renamer state. We do this becase
        # -- the Renamer class is able to utilise recollection to initialise
        # -- its state from its last run
        self.update_ui()

        # -- Ensure we update the ui whenever the renamer
        # -- is dynamically changed
        self.renamer.memento.stored.connect(
            self.update_ui,
        )

        # -- Hook up signals and slots for the buttons
        self.run_button.clicked.connect(self.renamer.run)
        self.browser_button.clicked.connect(self._set_directory)

        # -- Finally we hook up the events on our text fields so
        # -- the renamer object will recieve the users input
        # -- values
        text_propogation_widgets = dict(
            directory=self.directory_field,
            suffix_filter=self.suffix_filter,
            replace_this=self.replace_this,
            replace_with=self.with_this,
        )

        for label, text_widget in text_propogation_widgets.items():
            text_widget.textChanged.connect(
                partial(
                    self._push_change_to_renamer,
                    label,
                )
            )

        # -- Private variable to testing whether we actually need
        # -- to be propogating changes to the Renamer
        self._enable_propogation = True

    # --------------------------------------------------------------------------
    def _push_change_to_renamer(self, label, value):
        """
        This method pushes any text changes over to the renamer 
        object - allowing that object to carry out any recollection
        calls.
        
        :param label: The property name on the renamer to set 
        :param value: The value to set the property to
        
        :return: None 
        """
        setattr(
            self.renamer,
            label,
            value,
        )

    # --------------------------------------------------------------------------
    def _labelled_lineedit(self, label):
        """
        This is a layout setup method which makes it a little easier to 
        create a text field which a label in front of it.
        
        :param label: The text the label should show
         
        :return: QHBoxLayout, QLineEdit 
        """
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel(label))
        lineedit = QtWidgets.QLineEdit()
        layout.addWidget(lineedit)

        return layout, lineedit

    # --------------------------------------------------------------------------
    def _set_directory(self, directory=None):
        """
        Triggered whenever the user clicks the '...' button to select a 
        target directory. 
        :param directory: 
        :return: 
        """

        # -- If no directory is given we prompt for one
        if directory is None or directory is False:
            directory = QtWidgets.QFileDialog.getExistingDirectory(
                self,
                'Select directory to search from',
            )

        # -- If we still have no valid directory we exit
        if directory is None:
            return

        self.directory_field.setText(
            directory,
        )

    # --------------------------------------------------------------------------
    def update_ui(self):
        """
        This will read the data from the renamer and update the ui.
        
        :return: 
        """
        try:
            # -- We do not want to push changes back to the renamer
            # -- while we're updating our changes based on the renamers
            # -- current state, therefore we temporarily disable any
            # -- data propagation
            self._enable_propogation = False

            # -- Set our text fields
            self.with_this.setText(self.renamer.replace_with)
            self.replace_this.setText(self.renamer.replace_this)
            self.directory_field.setText(self.renamer.directory)
            self.suffix_filter.setText(self.renamer.suffix_filter)

            # -- Update our output log. Clear any previous results
            # -- then populate it based on the last run.
            self.log_output.clear()
            for success in self.renamer.processed:
                self.log_output.addItem(success)

            for failure in self.renamer.failed:
                self.log_output.addItem(failure)

        finally:
            # -- Now all our changes are complete we can turn data
            # -- propogation back on.
            self._enable_propogation = True


# ------------------------------------------------------------------------------
def demo():

    if not QtWidgets.QApplication.instance():
        QtWidgets.QApplication([])

    q_app = QtWidgets.QApplication.instance()
    widget = FileRenamerUi()
    widget.show()

    q_app.exec_()



import recollection
import unittest


# ------------------------------------------------------------------------------
class SimpleTestClass(recollection.Inference):

    def __init__(self):
        super(SimpleTestClass, self).__init__()
        self._letter = 1
        self.number = 10
        self.memento.register('number')

    @recollection.infer.get('letter')
    def letter(self):
        return self._letter

    @ recollection.infer.store('letter')
    def set_letter(self, value):
        self._letter = value


# ------------------------------------------------------------------------------
class InvalidTestClass(recollection.Inference):

    def __init__(self):
        super(InvalidTestClass, self).__init__()
        self._letter = 1

    @ recollection.infer.store('letter')
    def set_letter(self, value):
        self._letter = value


# ------------------------------------------------------------------------------
class TestInference(unittest.TestCase):

    # --------------------------------------------------------------------------
    def test_can_instance_an_inference_class(self):
        """
        Simply checks that an inference class can be instanced

        :return:
        """
        SimpleTestClass()

    # --------------------------------------------------------------------------
    def test_properties_are_stored_automatically(self):
        """
        Ensures that property changes are automatically stored

        :return:
        """
        test_class = SimpleTestClass()

        test_class.number = 99
        test_class.number = 42

        test_class.memento.restore(1)

        self.assertEqual(
            99,
            test_class.number,
        )

    # --------------------------------------------------------------------------
    def test_setters_are_triggered(self):
        """
        Ensures storing is triggered when decorated with the infer.store

        :return:
        """
        test_class = SimpleTestClass()

        test_class.set_letter('z')
        test_class.set_letter('x')
        test_class.memento.restore(1)

        self.assertEqual(
            'z',
            test_class.letter(),
        )

    # --------------------------------------------------------------------------
    def test_register_only_double_decorators(self):
        """
        Ensures that if you have incorrectly decorated a class an exception
        will be triggered.

        :return:
        """
        test_class = InvalidTestClass()

        self.assertEqual(
            [],
            test_class.memento.labels(),
        )

    # --------------------------------------------------------------------------
    def test_defer_changes(self):
        """
        Ensures that we do not get state changes during a deferred step

        :return:
        """
        # -- Create a class and define a property
        test_class = SimpleTestClass()

        with test_class.memento.defer():
            for i in range(10):
                test_class.number = i

        self.assertEqual(
            1,
            len(test_class.memento._states)
        )

    # --------------------------------------------------------------------------
    def test_omit_changes(self):
        """
        Ensures that we do not get state changes during a deferred step

        :return:
        """
        # -- Create a class and define a property
        test_class = SimpleTestClass()

        with test_class.memento.omit():
            for i in range(10):
                test_class.number = i

        self.assertEqual(
            0,
            len(test_class.memento._states)
        )



# Copyright (c) 2018 Michael Malinowski
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
"""
Recollection is a state recalling system. It allows for state of objects
to be snap-shotted exposing functionality to then 'rollback' to any
previously stored state.

Memento gives two distinctly different exposures of this mechanism. If 
performance is not the most critical concern and your focus is on code
simplicity then the inheritence approach is possibly your best option
providing you have ownership of the classes which need historical state.

This example shows how to setup a class using memento inheritence to 
automatically handle state storing. As you can see from the example, there
is no need to explicitely ask memento to store at any time as it is handled
entirely for you :

    ..code-block:: python
        
    >>> import recollection
    >>> 
    >>> # -- Inference is the Memento class designed specifically
    >>> # -- for inheritence situations.
    >>> class Foo(recollection.Inference):
    ... 
    ...     def __init__(self):
    ...         super(Foo, self).__init__()
    ... 
    ...         # -- Demonstrate a private attribute with getter
    ...         # -- and setter methods
    ...         self._letter = 1
    ... 
    ...         # -- Demonstrate public attributes which should be
    ...         # -- stored whenever they are changed
    ...         self.number = 10
    ... 
    ...         # -- Because we're inheriting from the Inference class, any
    ...         # -- registered properties will automatically trigger a
    ...         # -- state store whenever they are changed.
    ...         self.memento.register('number')
    ... 
    ...     # -- Declare that this is a memento getter
    ...     @recollection.infer.get('foobar')
    ...     def letter(self):
    ...         return self._letter
    ... 
    ...     # -- Declare this as a memento setter, using the same
    ...     # -- label as the getter. Whenever setHeight is called
    ...     # -- it will automatically store the state change.
    ...     @recollection.infer.store('foobar')
    ...     def setLetter(self, value):
    ...         self._letter = value
    >>> 
    >>> # -- Instance our object
    >>> foo = Foo()
    >>> 
    >>> # -- Update our variable using our accessor - which will automatically
    >>> # -- incur a state store
    >>> for letter in ['a', 'b', 'c', 'd', 'e']:
    ...     foo.setLetter(letter)
    >>> 
    >>> # -- Demonstrate that the current state is 'e'
    >>> print(foo.letter())  # -- Prints 'e'
    >>> 
    >>> # -- Roll back one setp and demonstrate that the property
    >>> # -- now evaluates to 'd'
    >>> foo.memento.restore(1)
    >>> print(foo.letter())  # -- Prints 'd'
    >>> 
    >>> # -- Here we demonstrate directly changing properties
    >>> # -- which are registered
    >>> foo.number = 5
    >>> foo.number = 99
    >>> print(foo.number == 99)
    >>> 
    >>> # -- Now restore back one step
    >>> foo.memento.restore(1)
    >>> print(foo.number == 5)

However, we do not always have the luxury of changing class inheritence
or you may specifically want to keep the memento state management out
of your actual inheeritence hierachy. The following examples all demonstrate
how this can be achieved.

In this example we have a class with two properties. We the instance
a Memento class targeting our foo instance. Each time we call the
store method within Memento we are taking a snapshot of the values
returned by the registered properties/functions

..code-block:: python

    >>> import recollection
    >>>
    >>>
    >>> class Foo(object):
    ...     def __init__(self):
    ...         self.name = 'bar'
    ...         self.i = 0
    >>>
    >>> # -- Instance our object
    >>> foo = Foo()
    >>>
    >>> # -- Instance a memento object pointing at foo
    >>> stack = recollection.Memento(foo)
    >>> stack.register('i')
    >>>
    >>> # -- Start changing some values on foo, and
    >>> # -- ask our stack to store those changes
    >>> for i in range(11):
    ...     foo.i = i
    ...
    ...     # -- Ask the memento object to store the state
    ...     stack.store()
    >>>
    >>> # -- Printing i, shows us 10
    >>> print(foo.i)
    >>>
    >>> # -- But lets say we roll back to the state 5 versions
    >>> # -- ago
    >>> stack.restore(5)
    >>>
    >>> # -- Now we can see i is at the version it was when
    >>> # -- it was stored 5 versions back
    >>> print(foo.i)

It also allows multiple Memento objects to be put into a lock-step,
such that whenever one memento object is storing or restoring then
all other memento objects in that sync group will also store or
restore.

..code-block:: python

    >>> import recollection
    >>>
    >>>
    >>> class Foo(object):
    ...     def __init__(self):
    ...         self.name = 'bar'
    ...         self.i = 0
    >>>
    >>> # -- This time we instance two completely seperate
    >>> # -- foo objects
    >>> foo_a = Foo()
    >>> foo_b = Foo()
    >>>
    >>> # -- Instance a memento stack for each
    >>> stack_a = recollection.Memento(foo_a)
    >>> stack_b = recollection.Memento(foo_b)
    >>>
    >>> stack_a.register(['name', 'i'])
    >>> stack_b.register(['name', 'i'])
    >>>
    >>> # -- Now we will put our stacks into a state of lock-step
    >>> # -- which means whenever one of them is stored or restored
    >>> # -- all others in the lock-step group will have the same
    >>> # -- action performed
    >>> stack_a.group(stack_b)
    >>>
    >>> # -- Increment some values on both objects
    >>> for i in range(11):
    ...     foo_a.i = i
    ...     foo_b.i = i
    ...
    ...     # -- Trigger a store on only one stack
    ...     stack_a.store()
    >>>
    >>> # -- We can see that both A and B have a value of 10
    >>> print(foo_a.i == 10 and foo_b.i == 10)
    >>>
    >>> # -- Now we rollback - knowing that this action will occur
    >>> # -- across all grouped memento objects
    >>> stack_a.restore(5)
    >>>
    >>> # -- Now we can see i is at the version it was when
    >>> # -- it was stored 5 versions back
    >>> print(foo_b.i)
    >>> print(foo_a.i == 5 and foo_b.i == 5)

Serialisers can also be registered against memento instances allowing
the stored state of a memento object to be serialised into a persistent
state.

This example shows how we might define a user preferences class, and
within that class we define a memento object to store the preference
state. By registering a serialiser the preferences state will be written
to disk whenever the 'store' is called.

Notice that in this example we're also choosing not to store private
member variables, but instead we're harnessing the public api of the
class as getters and setters.

..code-block:: python

    >>> class UserPreferences(object):
    ...
    ...     def __init__(self):
    ...
    ...         # -- Visual preferences
    ...         self._theme = 'default'
    ...
    ...         # -- Define our memento, which we utilise specifically to
    ...         # -- store our preferences to a persistent location
    ...         self._stack = recollection.Memento(self)
    ...
    ...         # -- We will utilise the JSON Appdata serialiser, which
    ...         # -- writes our memento information to the app data system
    ...         # -- location
    ...         self._stack.register_serialiser(
    ...             serialiser=recollection.JsonAppSerialiser,
    ...             identifier='memento/demos/userprefs/UserPreferenceA',
    ...         )
    ...
    ...         # -- Register which properties we want the store to focus on
    ...         self._stack.register(
    ...             label='theme',
    ...             getter=self.get_theme,
    ...             setter=self.set_theme,
    ...         )
    ...
    ...         # -- Finally, we deserialise - which will update this class
    ...         # -- with any previously stored state
    ...         self._stack.deserialise()
    ...
    ...     # --------------------------------------------------------------
    ...     def get_theme(self):
    ...         return self._theme
    ...
    ...     # --------------------------------------------------------------
    ...     def set_theme(self, theme):
    ...         self._theme = theme
    ...         self._stack.store(serialise=True)

Equally, if we want to make it a little more obvious at the class level 
which functions are storing we could opt to utilise the Memento decorator,
which stores and serialises:

    ..code-block:: python
    
    >>> from recollection import Memento
    >>>
    >>> class UserPreferences(object):
    ...
    ...     @Memento.serialise_after('theme')
    ...     def set_theme(self, theme):
    ...         self._theme = theme

Here we see an example of state being used as a roll-back feature, which
is particularly useful when allowing users to interact with data:

    ..code-block:: python

        >>> import recollection
        >>>
        >>>
        >>> class Foo(object):
        ...
        ...     def __init__(self):
        ...         self._x = 10
        ...         self._y = 20
        ...
        ...         self._stack = recollection.Memento(self)
        ...         self._stack.register(['_x', '_y',])
        ...
        ...         # -- Store the default state
        ...         self._stack.store()
        ...
        ...     def x(self):
        ...         return self._x
        ...
        ...     def setX(self, value):
        ...         self._x = value
        ...         self._stack.store()
        ...
        ...     def y(self):
        ...         return self._x
        ...
        ...     def setY(self, value):
        ...         self._y = value
        ...         self._stack.store()
        ...
        ...     def undo(self):
        ...         self._stack.restore()
        >>>
        >>> # -- Instance our foo object and print the default
        >>> # -- value
        >>> foo = Foo()
        >>> print('Default Value : %s' % foo.x())
        Default Value : 10
        >>>
        >>> # -- Set the value to 200
        >>> foo.setX(200)
        >>> print('Changed Value : %s' % foo.x())
        Changed Value : 200
        >>>
        >>> # -- Undo the last action, and print that we're not back
        >>> # -- to a value the same as the default
        >>> foo.undo()
        >>> print('Back to the default value after undo : %s' % foo.x())
        Back to the default value after undo : 10

These mechanics are demonstrated in the example modules, specifically:

    ..code-block:: python

        >>> # -- This demo shows a user preferences object being interacted
        >>> # -- which which works in the same way as the example above.
        >>> from recollection.examples.userprefs.demo import demo
        >>>
        >>> demo()

    ..code-block:: python

        >>> # -- This demo shows a user preferences object being interacted
        >>> # -- which which works by decorating setter properties
        >>> from recollection.examples.userprefs.demo import demo2
        >>>
        >>> demo2()

    ..code-block:: python

        >>> # -- This demo utilises a 'boardgame' style scenario where
        >>> # -- we're given two players and the desire to 'undo' the results
        >>> # -- of turns if they are not desirable!
        >>> from recollection.examples.boardgame.game import demo
        >>>
        >>> demo()
    
    ..code-block:: python

        >>> # -- This demo utilises a 'boardgame' style scenario where
        >>> # -- we're given two players and the desire to 'undo' the results
        >>> # -- of turns if they are not desirable!
        >>> from recollection.examples.pins.demo import demo
        >>>
        >>> demo()

"""
__author__ = "Michael Malinowski"
__copyright__ = "Copyright (C) 2018 Michael Malinowski"
__license__ = "MIT"
__version__ = "0.9.3"

from .core import (
    Memento,
)

from .inference import (
    Inference,
)

from .serialiser import (
    SerialiserBase,
)

from .serialisers import (
    PickleSerialiser,
    JsonAppSerialiser,
)

from . exceptions import (
    StorageError,
    RegistrationError,
)

from . import infer



"""
This namespace holds Inference decorators.
"""
from .inference import Inference


# --------------------------------------------------------------------------
def get(label):
    """
    This is an inference decorator which can be used to decorate
    a getter method.

    ..code-block:: python
        
        >>> import recollection
        >>>
        >>> class Foo(recollection.Inference):
        ... 
        ...     def __init__(self):
        ...         super(Foo, self).__init__()
        ... 
        ...         # -- Demonstrate a private attribute with getter
        ...         # -- and setter methods
        ...         self._letter = 1
        ... 
        ...     # -- Declare that this is a memento getter
        ...     @recollection.infer.get('foobar')
        ...     def get_letter(self):
        ...         return self._letter

    :param label: You must specify a label to be used when storing
        this variable. For the registration to be processed there 
        must be a correlating setter defined on the class with the same
        label.
    :type label: str

    :return: 
    """
    def inner_decor(func):
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        # -- Assign our attributes
        inner.label = label
        inner.is_memento_getter = True
        inner.is_memento_setter = False

        return inner

    return inner_decor


# --------------------------------------------------------------------------
def store(label, copy_value=True, serialise=False):
    """
    This is an inference decorator which can be used to decorate
    a getter method.

    ..code-block:: python
        
        >>> import recollection
        >>> 
        >>> class Foo(recollection.Inference):
        ... 
        ...     def __init__(self):
        ...         super(Foo, self).__init__()
        ... 
        ...         # -- Demonstrate a private attribute with getter
        ...         # -- and setter methods
        ...         self._letter = 1
        ... 
        ...     # -- Declare that this is a memento getter
        ...     @recollection.infer.store('letter', serialise=True)
        ...     def set_letter(self):
        ...         return self._letter
        >>> foo = Foo()
        
    :param label: You must specify a label to be used when storing
        this variable. For the registration to be processed there 
        must be a correlating setter defined on the class with the same
        label.
    :type label: str

    :param copy_value: Default is True, this defines whether the object
        being stored will be copied or referenced. Typically if this is
        likely to be left to True, however if you want a reference to an 
        object to be stored in history rather than copies of the object you
        should set this to False.
    :type copy_value: bool

    :param serialise: If true this will perform a serialisation of the 
        memento object each time the setter is called. This expects a
        serialiser to be registered. The default is False.
    :type serialise: bool

    :return: 
    """

    def inner_decor(func):
        def inner(*args, **kwargs):
            result = func(*args, **kwargs)

            # -- Now store
            if isinstance(args[0], Inference):
                args[0].memento.store(serialise=serialise)

            return result

        # -- Assign our attributes
        inner.label = label
        inner.is_memento_getter = False
        inner.is_memento_setter = True
        inner.copy_value = copy_value

        return inner

    return inner_decor



from recollection.tests.classes import SetterTestClass

import os
import recollection
import tempfile
import unittest


# ------------------------------------------------------------------------------
class TestClass(object):
    # noinspection PyUnusedLocal
    def __init__(self):
        foo = 1
        bar = ''


# ------------------------------------------------------------------------------
class TestSerialiserAppData(unittest.TestCase):

    # --------------------------------------------------------------------------
    def test_serialise(self):
        """
        Checks to see that our serialisation kicks in and we have
        a persistent data artifact

        :return:
        """
        # -- Create out test stack
        test_class, stack = self._memento_test_data()

        # -- Define the data location
        temp_file = tempfile.NamedTemporaryFile(delete=True)
        temp_file.close()

        # -- Ensure the data exists
        self.assertFalse(
            os.path.exists(temp_file.name)
        )

        # -- Register our serialiser
        stack.register_serialiser(
            serialiser=recollection.PickleSerialiser,
            identifier=temp_file.name,
        )

        # -- Set some properties on the class
        test_class.setTarget(10)
        test_class.setDistance(20)

        # -- Store the data and serialise
        stack.store()
        stack.serialise()

        # -- Ensure the data exists
        self.assertTrue(
            os.path.exists(temp_file.name)
        )

    # --------------------------------------------------------------------------
    def test_serialise_on_store(self):
        """
        Checks to see that our serialisation kicks in and we have
        a persistent data artifact when calling the shortcut of
        serelialise=True on the store method

        :return:
        """

        # -- Create out test stack
        test_class, stack = self._memento_test_data()

        # -- Define the data location
        temp_file = tempfile.NamedTemporaryFile(delete=True)
        temp_file.close()

        # -- Ensure the data exists
        self.assertFalse(
            os.path.exists(temp_file.name)
        )

        # -- Register our serialiser
        stack.register_serialiser(
            serialiser=recollection.PickleSerialiser,
            identifier=temp_file.name,
        )

        # -- Set some properties on the class
        test_class.setTarget(10)
        test_class.setDistance(20)

        # -- Store the data and serialise
        stack.store(serialise=True)

        # -- Ensure the data exists
        self.assertTrue(
            os.path.exists(temp_file.name)
        )

    # --------------------------------------------------------------------------
    def test_deserialise(self):
        """
        Checks to see that our serialisation kicks in and we have
        a persistent data artifact when calling the shortcut of
        serelialise=True on the store method

        :return:
        """
        # -- Create out test stack
        test_class, stack = self._memento_test_data()

        # -- Define the data location
        temp_file = tempfile.NamedTemporaryFile(delete=True)
        temp_file.close()

        # -- Ensure the data exists
        self.assertFalse(
            os.path.exists(temp_file.name)
        )

        # -- Register our serialiser
        stack.register_serialiser(
            serialiser=recollection.PickleSerialiser,
            identifier=temp_file.name,
        )

        # -- Set some properties on the class
        test_class.setTarget(10)
        test_class.setDistance(20)

        # -- Store the data and serialise
        stack.store(serialise=True)

        # -- Create out test stack
        new_test_class, new_stack = self._memento_test_data()

        # -- Register our serialiser
        new_stack.register_serialiser(
            serialiser=recollection.PickleSerialiser,
            identifier=temp_file.name,
        )
        new_stack.deserialise()

        # -- Ensure the data exists
        self.assertEqual(
            10,
            new_test_class.getTarget(),
        )

        self.assertEqual(
            20,
            test_class.getDistance(),
        )

    # --------------------------------------------------------------------------
    @classmethod
    def _memento_test_data(cls):

        test_class = SetterTestClass()

        stack = recollection.Memento(test_class)
        stack.register(
            label='target',
            getter=test_class.getTarget,
            setter=test_class.setTarget,
        )

        stack.register(
            label='distance',
            getter=test_class.getDistance,
            setter=test_class.setDistance,
        )

        return test_class, stack



"""
Compatibility code for py2/py3
"""

# Naming convention: Variables named "wr" are weak reference objects;
# they are called this instead of "ref" to avoid name collisions with
# the module-global ref() function imported from _weakref.

from weakref import (
     ref,
)


class WeakMethod(ref):
    """
    A custom `weakref.ref` subclass which simulates a weak reference to
    a bound method, working around the lifetime problem of bound methods.
    """

    __slots__ = "_func_ref", "_meth_type", "_alive", "__weakref__"

    def __new__(cls, meth, callback=None):
        try:
            obj = meth.__self__
            func = meth.__func__
        except AttributeError:
            raise TypeError(
                "argument should be a bound method, not {}".format(type(meth)),
            )

        def _cb(arg):
            # The self-weakref trick is needed to avoid creating a reference
            # cycle.
            self = self_wr()
            if self._alive:
                self._alive = False
                if callback is not None:
                    callback(self)
        self = ref.__new__(cls, obj, _cb)
        self._func_ref = ref(func, _cb)
        self._meth_type = type(meth)
        self._alive = True
        self_wr = ref(self)
        return self

    def __call__(self):
        obj = super(WeakMethod, self).__call__()
        func = self._func_ref()
        if obj is None or func is None:
            return None
        return self._meth_type(func, obj)

    def __eq__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is other
            return ref.__eq__(self, other) and self._func_ref == other._func_ref
        return False

    def __ne__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is not other
            return ref.__ne__(self, other) or self._func_ref != other._func_ref
        return True

    __hash__ = ref.__hash__



from ..serialiser import SerialiserBase

import os

try:
    # noinspection PyPep8Naming
    import cPickle as pickle

except ImportError:
    import pickle


# ------------------------------------------------------------------------------
class PickleSerialiser(SerialiserBase):
    """
    This will serialise the memento object to a json file and store it in
    an application data location (varies depending on platform). You can
    utilise the identifier as a sub-path to create namespaced serialisation
    objects.

    For instance:
        serialiser.serialise(memento, 'foo/bar'

    would (on windows) result in:
        %APPDATA%/memento/foo/bar.json

    """

    # --------------------------------------------------------------------------
    @staticmethod
    def serialise(data, identifier):
        """
        The identifier for this serialiser should be the absolute path
        to the file you want to write to
        
        :param data: dict
        :param identifier: str
         
        :return: 
        """
        directory = os.path.dirname(identifier)

        if not os.path.exists(directory):
            os.makedirs(directory)

        with open(identifier, 'wb') as f:
            pickle.dump(
                data,
                f,
            )

    # --------------------------------------------------------------------------
    @staticmethod
    def deserialise(identifier):
        """
        The identifier for this serialiser should be the absolute path
        to the file you want to write to

        :param identifier: str
         
        :return: 
        """
        if not os.path.exists(identifier):
            return None

        with open(identifier, 'rb') as f:
            return pickle.load(f)

    # --------------------------------------------------------------------------
    @staticmethod
    def locator(identifier):
        """
        With the given identifier this should return the resolved location
        of the data

        :param identifier: This is very dependent on the requirements of
            each individual serialiser. This could be a filepath, a url,
            a name etc.
        :type identifier: str

        :return: str
        """
        return identifier



import random
import recollection


# ------------------------------------------------------------------------------
def demo():
    """
    This demo creates a preferences class which calls the store on
    the setters for the properties we want to store. It then instances
    a second class to check that the preferences were deserialised
    correctly.

    :return:
    """
    # -- Instance our user prefs
    pref_a = UserPreferencesA()

    # -- Set a preference
    new_theme = random.choice(UserPreferencesA.THEMES)
    pref_a.theme = new_theme

    # -- Now lets create another instance of the same preferences
    # -- and check that the preference was both stored and restored
    pref_a_2 = UserPreferencesA()

    print(
        'User Preferences are : %s, expected was : %s' % (
            pref_a_2.theme,
            new_theme,
        )
    )


# ------------------------------------------------------------------------------
def demo2():
    """
    This demo creates a preferences class which is decorated on the
    setters to perform a serialisation on call. It then instances
    a second class to check that the preferences were deserialised
    correctly.

    :return:
    """
    # -- Instance our user prefs
    pref_a = UserPreferencesB()

    # -- Set a preference
    new_theme = random.choice(UserPreferencesB.THEMES)
    pref_a.theme = new_theme

    # -- Now lets create another instance of the same preferences
    # -- and check that the preference was both stored and restored
    pref_a_2 = UserPreferencesB()

    print(
        'User Preferences are : %s, expected was : %s' % (
            pref_a_2.theme,
            new_theme,
        )
    )


# ------------------------------------------------------------------------------
class UserPreferencesA(object):
    """
    This is some hypothetical user preferences for a piece of
    novel writing software.

    In this particular example we expose our preferences through
    getter and setter methods and ask our memento to store during
    the setting
    """

    THEMES = [
        'Red',
        'Green',
        'Blue',
        'Orange',
        'Tropical',
        'Ice',
        'Fire',
        'Scary',
    ]

    # --------------------------------------------------------------------------
    def __init__(self):

        # -- Visual preferences
        self._theme = 'default'
        self._show_spelling_errors = True

        # -- Save/loading preferences
        self._save_on_close = True
        self._load_last_project = True

        # -- Define our memento, which we utilise specifically to
        # -- store our preferences to a persistent location
        self._store = recollection.Memento(self)

        # -- We will utilise the JSON Appdata serialiser, which
        # -- writes our memento information to the app data system
        # -- location
        self._store.register_serialiser(
            serialiser=recollection.JsonAppSerialiser,
            identifier='memento/demos/userprefs/UserPreferenceA',
        )

        # -- Register which properties we want the store to focus on
        self._store.register(
            [
                '_theme',
                '_save_on_close',
                '_load_last_project',
                '_show_spelling_errors',
            ],
        )

        # -- Finally, we deserialise - which will update this class with
        # -- any previously stored state
        self._store.deserialise()

    # --------------------------------------------------------------------------
    @property
    def theme(self):
        return self._theme

    # --------------------------------------------------------------------------
    @theme.setter
    def theme(self, theme):
        self._theme = theme
        self._store.store(serialise=True)

    # --------------------------------------------------------------------------
    @property
    def show_spelling_errors(self):
        return self._show_spelling_errors

    # --------------------------------------------------------------------------
    @show_spelling_errors.setter
    def show_spelling_errors(self, value):
        self._show_spelling_errors = value
        self._store.store(serialise=True)

    # --------------------------------------------------------------------------
    @property
    def save_on_close(self):
        return self._save_on_close

    # --------------------------------------------------------------------------
    @save_on_close.setter
    def save_on_close(self, value):
        self._save_on_close = value
        self._store.store(serialise=True)

    # --------------------------------------------------------------------------
    @property
    def load_last_project(self):
        return self._load_last_project

    # --------------------------------------------------------------------------
    @load_last_project.setter
    def load_last_project(self, value):
        self._load_last_project = value
        self._store.store(serialise=True)


# ------------------------------------------------------------------------------
class UserPreferencesB(object):
    """
    This is some hypothetical user preferences for a piece of
    novel writing software.

    In this particular example we expose our preferences through
    getter and setter methods and ask our memento to store during
    the setting
    """

    THEMES = [
        'Red',
        'Green',
        'Blue',
        'Orange',
        'Tropical',
        'Ice',
        'Fire',
        'Scary',
    ]

    # --------------------------------------------------------------------------
    def __init__(self):

        # -- Visual preferences
        self._theme = 'default'
        self._show_spelling_errors = True

        # -- Save/loading preferences
        self._save_on_close = True
        self._load_last_project = True

        # -- Define our memento, which we utilise specifically to
        # -- store our preferences to a persistent location
        self._store = recollection.Memento(self)

        # -- We will utilise the JSON Appdata serialiser, which
        # -- writes our memento information to the app data system
        # -- location
        self._store.register_serialiser(
            serialiser=recollection.JsonAppSerialiser,
            identifier='memento/demos/userprefs/UserPreferenceA',
        )

        # -- Register which properties we want the store to focus on
        self._store.register(
            [
                '_theme',
                '_save_on_close',
                '_load_last_project',
                '_show_spelling_errors',
            ],
        )

        # -- Finally, we deserialise - which will update this class with
        # -- any previously stored state
        self._store.deserialise()

    # --------------------------------------------------------------------------
    @property
    def theme(self):
        return self._theme

    # --------------------------------------------------------------------------
    @theme.setter
    @recollection.Memento.serialise_after('_store')
    def theme(self, theme):
        self._theme = theme

    # --------------------------------------------------------------------------
    @property
    def show_spelling_errors(self):
        return self._show_spelling_errors

    # --------------------------------------------------------------------------
    @show_spelling_errors.setter
    @recollection.Memento.serialise_after('_store')
    def show_spelling_errors(self, value):
        self._show_spelling_errors = value

    # --------------------------------------------------------------------------
    @property
    def save_on_close(self):
        return self._save_on_close

    # --------------------------------------------------------------------------
    @save_on_close.setter
    @recollection.Memento.serialise_after('_store')
    def save_on_close(self, value):
        self._save_on_close = value

    # --------------------------------------------------------------------------
    @property
    def load_last_project(self):
        return self._load_last_project

    # --------------------------------------------------------------------------
    @load_last_project.setter
    @recollection.Memento.serialise_after('_store')
    def load_last_project(self, value):
        self._load_last_project = value


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    demo()
    demo2()



from .core import Memento
from .exceptions import StorageError


# ------------------------------------------------------------------------------
class Inference(object):
    """
    This class is specifically designed to allow you to inherit from
    it. In doing so the class will infer storage and serialisation behaviour
    for you through declared decorators.
    
    The other additional benefit of using this mechanism is that it can 
    monitor and store changes automatically whenever registered properties
    are changed.
    
    Due to inference testing during calls this approach is not as performant
    as creating a standard Memento object and explicitely calling the store
    methods where required. However, if performance is not the key concern 
    this class can make the code implementation significantly lighter and
    easier to manage.
    
    This example shows how to mark public attributes for automatic storage
    as well as defining setters and getters:
    
    ..code-block:: python
        
        >>> import recollection
        >>> 
        >>> class Foo(recollection.Inference):
        ... 
        ...     def __init__(self):
        ...         super(Foo, self).__init__()
        ... 
        ...         # -- Demonstrate a private attribute with getter
        ...         # -- and setter methods
        ...         self._letter = 1
        ... 
        ...         # -- Demonstrate public attributes which should be
        ...         # -- stored whenever they are changed
        ...         self.number = 10
        ... 
        ...         # -- Because we're inheriting from the Inference class, any
        ...         # -- registered properties will automatically trigger a
        ...         # -- state store whenever they are changed.
        ...         self.memento.register('number')
        ... 
        ...     # -- Declare that this is a memento getter
        ...     @recollection.infer.get('foobar')
        ...     def letter(self):
        ...         return self._letter
        ... 
        ...     # -- Declare this as a memento setter, using the same
        ...     # -- label as the getter. Whenever setHeight is called
        ...     # -- it will automatically store the state change.
        ...     @recollection.infer.store('foobar')
        ...     def setLetter(self, value):
        ...         self._letter = value
        >>> 
        >>> # -- Instance our object
        >>> foo = Foo()
        >>> 
        >>> # -- Update our variable using our accessor - which will automatically
        >>> # -- incur a state store
        >>> for letter in ['a', 'b', 'c', 'd', 'e']:
        ...     foo.setLetter(letter)
        >>> 
        >>> # -- Demonstrate that the current state is 'e'
        >>> print(foo.letter())  # -- Prints 'e'
        >>> 
        >>> # -- Roll back one setp and demonstrate that the property
        >>> # -- now evaluates to 'd'
        >>> foo.memento.restore(1)
        >>> print(foo.letter())  # -- Prints 'd'
        >>> 
        >>> # -- Here we demonstrate directly changing properties
        >>> # -- which are registered
        >>> foo.number = 5
        >>> foo.number = 99
        >>> print(foo.number == 99)
        >>> 
        >>> # -- Now restore back one step
        >>> foo.memento.restore(1)
        >>> print(foo.number == 5)
    """

    # ----------------------------------------------------------------------
    def __init__(self):

        # -- Define the memento object we will use for
        # -- state storage
        self.memento = Memento(self)

        self._initialise_decorator_registration()

    # --------------------------------------------------------------------------
    def __getattribute__(self, item):
        return object.__getattribute__(self, item)

    # --------------------------------------------------------------------------
    def __setattr__(self, name, value):
        """
        We re-implement the setattr to allow us to perform a check as to
        whether the name we're trying to set is registered in the list of
        watch items in the Memento object.
        
        :param name: 
        :param value: 
        :return: 
        """
        self.__dict__[name] = value

        try:
            if name in self.memento.labels():
                self.memento.store()

        except (AttributeError, StorageError):
            pass

    # --------------------------------------------------------------------------
    def _initialise_decorator_registration(self):
        """
        This will perform an intial registration based on the decorators
        wrapping any methods of this class.

        :return: 
        """
        # -- We need to
        found_getters = dict()
        found_setters = dict()

        # -- Cycle over all the elements making up the class
        # -- and inspect them for decorators
        for item in dir(self):

            # -- Get the item as its object
            item = getattr(
                self,
                item,
            )

            # -- Ask for forgiveness rather than permission - if the attributes
            # -- dont exist we just skip over to the next time.
            try:
                if item.is_memento_getter:
                    found_getters[item.label] = item

                elif item.is_memento_setter:
                    found_setters[item.label] = item

            except AttributeError:
                pass

        # -- For each found label, register it in the memento
        # -- object unless it is already registered
        for label in found_getters:

            # -- We only allow matching getters and setters
            if label not in found_setters:
                continue

            # -- If this is already registered then we do not
            # - need to register it again
            if label in self.memento.labels():
                continue

            # -- Do the final registration
            self.memento.register(
                label=label,
                getter=found_getters[label],
                setter=found_setters[label],
                copy_value=False
            )



from . import exceptions
from .constants import log
from .signal import Signal
from .serialiser import SerialiserBase
from contextlib import contextmanager

import six
import copy
import weakref


# ------------------------------------------------------------------------------
class Memento(object):
    """
    The Memento class aims to lower the development overhead of implementing
    storable state as well as offering functionality to step backward to 
    previous states along with state serialisation for persistent state.
    """

    # -- This is a list of recollection objects which should all store
    # -- and restore whenever any other in the sync group does.
    _SYNC_GROUPS = list()

    # --------------------------------------------------------------------------
    def __init__(self, target, max_states=100):

        # -- Define our callback signals to allow other mechanisms
        # -- to hook into recollection events
        self.stored = Signal()
        self.restored = Signal()
        self.registered = Signal()
        self.unregistered = Signal()

        # -- If True, this property will omit any calls
        # -- to store. This is a useful mechanism when wanting
        # -- to defer lots of changes into a single step
        self._defer = False

        # -- Store the object we want to store state
        # -- for
        self._target = weakref.ref(target)

        # -- Define the list which we will store states in
        self._states = list()
        self._max_states = max_states

        # -- Store properties need to read when serialising
        # -- and de-serialising
        self._items_to_record = list()

        # -- Store the registered serialiser (if required)
        self._serialiser = None
        self._serialisation_identifier = None
        self._always_serialise = False

    # --------------------------------------------------------------------------
    def register(self, getter, setter=None, label=None, copy_value=True):
        """
        This registers an attribute/method to call or read whenever the 
        store method is called.
        
        :param getter: This is what should be used to get the value.
        :type getter: If a string, the object the recollection is trying to 
            store state for will be accessed using a 'getattr' call. If
            a method or function the given function will be stored - and
            is expected to be callable with no arguments.
            Alternatively this can be a list, in which case this function
            will be called for each element in the list.
            
        :param setter: If omitted the getter is used for the setter. This 
            is typical for attributes - where we do not need to need to set
            any differently to how we get. However, if you have methods for 
            setting and getting (such as getName, setName) you can define
            the setter here.
        :type setter: str or method
        
        :param label: This allows you to tailor how the values relating
            to this registration are stored. If the getter is a string this
            argument can be omitted, otherwise it should be provided.
        :type label: str
        
        :param copy_value: This is important if you're storing complex object
            types which are mutable. By default we take a copy of all values
            we recieve to ensure they cannot be modified outside of the 
            snapshot. However, if you want to store classes where you 
            specifically do not want them copied and instead want them
            referenced you may set this value to false.
        :type copy_value: bool
        :return: 
        """

        # -- If we're given a list we consider it a list of argument
        # -- sets, so we call this function for each element.
        if isinstance(getter, (list, tuple)):
            for item in getter:
                if isinstance(item, (list, tuple)):
                    self.register(*item)

                else:
                    self.register(item)

            return

        # -- If a setter was not defined we assume we're
        # -- to use the getter for both
        setter = setter or getter
        label = label or getter

        # -- Get a list of existing labels so we can check for dupilcates
        existing_labels = [
            record_.label
            for record_ in self._items_to_record
        ]

        # -- Do a final test to ensure we have everything we need
        if label in existing_labels:
            raise exceptions.RegistrationError(
                '%s is already registered. You must un-register this entry '
                'before re-registering.'
            )

        # -- Create an accessing pair
        item = _ItemToRecord(
            target=self._target,
            label=label,
            getter=getter,
            setter=setter,
            copy_value=copy_value
        )

        # -- Store the entries
        self._items_to_record.append(item)

        # -- Emit the event
        self.registered.emit()

        # -- Log the result
        log.debug(
            'Registered %s : \n\tGetter: %s \n\tSetter: %s' % (
                label,
                getter,
                setter,
            )
        )

    # --------------------------------------------------------------------------
    def unregister(self, label):
        """
        This will unregister an item from recollection. By unregistering an item
        that item will not longer be stored in any subsequent sates. It will
        not remove it from previous states, but it will be ignored data (in that
        unregistered data will never be restored).
        
        :param label:  This should be the label (or name of the attribute)
            to be unregistered
        :type label: str
        :return: 
        """
        # -- To prevent list mutation during iteration
        # -- we get a list of items we need to remove
        items_to_remove = [
            item
            for item in self._items_to_record
            if item.label == label
        ]

        # -- Iterate until all these items are gone
        while items_to_remove:
            item = items_to_remove.pop()

            self._items_to_record.remove(item)

            # -- Emit the event
            self.unregistered.emit()

            # -- Log the removal
            log.debug(str(item))

    # --------------------------------------------------------------------------
    def store(self, serialise=False):
        """
        This will initiate a snapshot of the targets values, calling or 
        retrieving values from any of the registered attributes or methods.
        
        :param serialise: If true a serialisation (to persistent data) will
            be performed once the store is complete. If no serialiser is 
            registered this will be be skipped.
        :type serialise: bool 
        """
        if self._defer:
            return

        # -- Given that our target is a weak ref we need to access
        # -- it through a call. If it has already been garbage
        # -- collected we need to raise an exception
        if not self._target():
            raise exceptions.StorageError('Target object is no longer in scope')

        # -- Define the dictionary to which we will store all the
        # -- targets details
        snapshot = dict()

        # -- We specifically store information based on what has been
        # -- registered wtih the recollection object, so we cycle that
        for item in self._items_to_record:
            snapshot[item.label] = item.get()

        # -- Insert the new stored state at the start
        # -- of the list
        self._states.insert(0, snapshot)

        # -- If our max states has been exceeded we start
        # -- removing them
        if self._max_states and len(self._states) > self._max_states:
            self._states.pop()

        # -- If we need to serialise, do so now
        if serialise or self._always_serialise:
            self.serialise()

        # -- Add some debug output
        log.debug('Stored Snapshot for %s' % self._target)

        # -- Emit the event
        self.stored.emit()

        # -- Call the store on any members this is marked as keeping
        # -- in sync step with
        with _SyncGroupPropogation(self) as sync_group:
            for memento in sync_group:
                memento.store(serialise=serialise)

    # --------------------------------------------------------------------------
    def restore(self, index=0):
        """
        This will restore the state of the target object to that of the
        given index. 
        
        An index of 0 is the current state - and therefore will have no 
        impact. An index of 1 will mean the state will go back 1 step. An
        index of 2 will mean the state will go back two steps and so forth.
        
        :param index: The amount of steps to go back by.
        :type index: int 
        """

        with self.defer():
            # -- Access the snapshot from the state stack
            snapshot = self._states[index]

            # -- Cycle over all the items we have been told to record
            for item in self._items_to_record:
                item.set(snapshot[item.label])

            # -- Log the removal
            log.debug(
                'Restored State [%s] to : %s' % (
                    index,
                    snapshot,
                )
            )

            # -- Call the store on any members this is marked as keeping
            # -- in sync step with
            with _SyncGroupPropogation(self) as sync_group:
                for memento in sync_group:
                    memento.restore(index=index)

        # -- Emit the event
        self.restored.emit()

    # --------------------------------------------------------------------------
    @contextmanager
    def defer(self, serialise=False):
        """
        This allows the deferring of stores for the duration of the context
        after which a store is called.
        
        :param serialise: Whether to serialise once the context finishes
        
        :return: None
        """
        self._defer = True
        yield None
        self._defer = False
        self.store(serialise=serialise)

    # --------------------------------------------------------------------------
    @contextmanager
    def omit(self):
        """
        This allows changes to be made without any storing occuring even
        though decorated functions or registered properties may be altered.

        :return: None
        """
        self._defer = True
        yield None
        self._defer = False

    # --------------------------------------------------------------------------
    def register_serialiser(self,
                            serialiser,
                            identifier,
                            always_serialise=False):
        """
        This allows you to register the serialiser you want to utilise.

        :param serialiser: Serialiser type. This does not need to be
            instanced, as all methods on serialisers are static.
        :type serialiser: Type recollection.SerialiserBase

        :param identifier: This is the identifier that should be given
            to the serialiser in order to process the request and store
            the data in the expected way. You should read the documentation
            of the serialiser in question to find the specific expectations.
        :type identifier: str
        
        :param always_serialise: If true, this will always serialise regardless
            of the argument given during the store call.
        :type always_serialise: bool
        
        :return: None
        """
        # -- Validate that the serialiser is indeed a serialiser!
        if not issubclass(serialiser, SerialiserBase):
            raise TypeError(
                '%s is not a subclass of %s' % (
                    serialiser,
                    SerialiserBase,
                )
            )

        # -- Store both the serialiser and the identifer at the
        # -- instance level to make it nice and easy for the
        # -- user to call a serialisation process.
        self._serialiser = serialiser
        self._serialisation_identifier = identifier
        self._always_serialise = always_serialise

        # -- Log the removal
        log.debug(
            'Serialiser Registered : %s' % serialiser.__class__.__name__,
        )

    # --------------------------------------------------------------------------
    def unregister_serialiser(self):
        """
        This will unregister the serialiser. If the serialiser was set to 
        always serialise this option will be set back to False.
        
        :return: None 
        """
        self._serialiser = None
        self._always_serialise = False
        self._serialisation_identifier = ''

    # --------------------------------------------------------------------------
    def labels(self):
        return [
            item.label
            for item in self._items_to_record
        ]

    # --------------------------------------------------------------------------
    def count(self):
        """
        Returns how many states have currently been stored
        :return: 
        """
        return len(self._states)

    # --------------------------------------------------------------------------
    def group(self, other):
        """
        Occasionally you many want to instantiate multiple recollection objects
        but have them all store and restore in lock-step. This can be done
        by grouping recollection objects together. 
        
        :param other: The recollection object you want to put into lock-step  
            with this group.
            NOTE: If the 'other' group is already in lock-step with a third
            group then all three groups will go into lock-step together.
        :type other: memento.Memento 
        """
        # -- Cycle over the sync groups, if we find a sync group
        # -- which contains this recollection we add the other to the
        # -- same group
        for idx in range(len(Memento._SYNC_GROUPS)):
            if self in Memento._SYNC_GROUPS[idx]:
                if other not in Memento._SYNC_GROUPS[idx]:
                    Memento._SYNC_GROUPS[idx].append(other)
                return

        # -- To get here we need to define a new sync group
        Memento._SYNC_GROUPS.append(
            [
                self,
                other,
            ],
        )

        # -- Log the removal
        log.debug(
            '%s added to group along with : %s' % (
                self,
                self.sync_group(),
            )
        )

    # --------------------------------------------------------------------------
    def sync_group(self):
        """
        This will return all the members of the lock-step. If there is no
        grouping the recollection will be returning in its own list/group.
         
        :return: list(mement.Memento, ...) 
        """
        for group in Memento._SYNC_GROUPS:
            if self in group:
                return list(group)

        # -- If there is no group, we consider it to be alone
        # -- in its own group
        return [self]

    # --------------------------------------------------------------------------
    def serialise(self):
        """
        Providing a serialiser has been registered, this will intiate a 
        serialisation process - allowing this recollection's current state to 
        be stored in a persistent way. 
        """
        # -- If we have no serialiser but we are being requested
        # -- to serialise we raise an exception
        if not self._serialiser:
            raise Exception(
                'No serialiser has been defined for this recollection object',
            )

        # -- Ask the registered serialise to serialise our latest
        # -- state
        self._serialiser.serialise(
            self._states[0],
            self._serialisation_identifier,
        )

        # -- Log the removal
        log.debug('Serialised State on %s' % self)

    # --------------------------------------------------------------------------
    def deserialise(self):
        """
        This will initialise the target with the persistently stored
        state (if available). 
        """
        # -- If we have no serialiser but we are being requested
        # -- to serialise we raise an exception
        if not self._serialiser:
            raise Exception(
                'No serialiser has been defined for this recollection object',
            )

        # -- Get the data from the deserialisation process
        deserialisation = self._serialiser.deserialise(
            self._serialisation_identifier,
        )

        # -- Providing that process was successful we add the returned
        # -- state into recollection and restore to it
        if deserialisation:
            self._states.insert(0, deserialisation)
            self.restore(index=0)

        log.debug('Deserialised State to %s' % self)

    # --------------------------------------------------------------------------
    @staticmethod
    def serialise_after(memento_accessor):
        """
        This decorator will initialise a serialisation of the
        data store whenever this function is called. This is the
        equivalent of calling memento.store(serialise=True) at
        the end of the function - but decorating can make it a
        little clearer at-a-glance which functions cause a state
        serialisation

        This should only be applied to instance methods of classes as it
        is always assumed that the first argument is the instance of the
        class.

        In the following example we set up our memento object and define
        a serialiser, but rather than call the memento store from within the
        setter we simply decorate the function. This gives no functional
        difference, but makes it easier to see at a definition level which
        methods will incur a serialisation

        ..code-block:: python
            
            >>> from recollection import Memento, JsonAppSerialiser
            >>>
            >>> class UserPreferencesB(object):
            ...
            ...     def __init__(self):
            ...
            ...         # -- Visual preferences
            ...         self._theme = 'default'
            ...
            ...         # -- Define our memento, which we utilise specifically 
            ...         # -- to store our preferences to a persistent location
            ...         self._store = Memento(self)
            ...
            ...         # -- We will utilise the JSON Appdata serialiser, which
            ...         # -- writes our memento information to the app data 
            ...         # -- system
            ...         # -- location
            ...         self._store.register_serialiser(
            ...             serialiser=JsonAppSerialiser,
            ...             identifier='memento/demos/prefs/UserPreferenceA',
            ...         )
            ...
            ...         # -- Register which properties we want the store to 
            ...         # -- focus on
            ...         self._store.register('_theme')
            ...
            ...         # -- Finally, we deserialise - which will update this 
            ...         # -- class with any previously stored state
            ...         self._store.deserialise()
            ...
            ...     @property
            ...     def theme(self):
            ...         return self._theme
            ...
            ...     @theme.setter
            ...     @Memento.serialise_after('_store')
            ...     def theme(self, theme):
            ...         self._theme = theme

        :param memento_accessor: This is the attribute name on the class
            to access

        :return:
        """

        # ----------------------------------------------------------------------
        def serialise_after_inner(func):

            # ------------------------------------------------------------------
            def inner(*args, **kwargs):

                # -- Call the function
                result = func(*args, **kwargs)

                try:
                    getattr(
                        args[0],
                        memento_accessor,
                    ).store(serialise=True)

                except AttributeError:
                    log.warning(
                        'Could not store state after call to %s' % func,
                    )

                return result

            return inner

        return serialise_after_inner


# ------------------------------------------------------------------------------
class _SyncGroupPropogation(object):
    """
    This is an internal and private class which manages recusion around
    lock-step grouping. This ensures that a sequencial group of recollection's
    will not continually call one-another when performing a lock-step call.
    """

    _ACTIVELY_PROCESSING = list()

    # --------------------------------------------------------------------------
    def __init__(self, memento):
        self._memento = memento
        self._processing_in_context = list()

    # --------------------------------------------------------------------------
    def __enter__(self):

        # -- If our recollection instance is already being processes then
        # -- there is nothing more to do and we return an empty
        # -- list
        if self._memento in _SyncGroupPropogation._ACTIVELY_PROCESSING:
            return list()

        # -- Get a list of members to actively process
        self._processing_in_context = self._memento.sync_group()

        # -- Add in the list of things we need to process, to ensure
        # -- they will not get re-processed
        _SyncGroupPropogation._ACTIVELY_PROCESSING.extend(
            self._processing_in_context
        )

        # -- Return a list of all the recollection's apart from this one
        return [
            memento
            for memento in self._processing_in_context
            if memento != self._memento
        ]

    # --------------------------------------------------------------------------
    def __exit__(self, exc_type, exc_val, exc_tb):
        # -- Now that we are done processing we need to remove any
        # -- instances out of the class list.
        while self._processing_in_context:
            self._ACTIVELY_PROCESSING.remove(
                self._processing_in_context.pop(),
            )


# ------------------------------------------------------------------------------
class _ItemToRecord(object):
    """
    This stores all the data required to be able to quickly and efficiently
    call or access data.
    """

    # --------------------------------------------------------------------------
    def __init__(self, target, label, getter, setter, copy_value):

        self.label = label

        self._target = target
        self._getter = getter
        self._setter = setter
        self._copy_value = copy_value

        self._label = self.label or getter

        self._get_is_callable = callable(self._getter)
        self._set_is_callable = callable(self._setter)

        if not isinstance(self.label, six.string_types):
            raise TypeError(
                'When giving a partial or method you must specify a '
                'label as it cannot be dynamically inferred'
            )

    # --------------------------------------------------------------------------
    def __repr__(self):
        return '[Unregistered %s, Getter: %s, Setter: %s]' % (
            self.label,
            self._getter,
            self._setter,
        )

    # --------------------------------------------------------------------------
    def _copy(self, item):
        if self._copy_value:
            return copy.deepcopy(item)
        return item

    # --------------------------------------------------------------------------
    def get(self):
        if self._get_is_callable:
            # -- We use a deepcopy to ensure that we're not storing
            # -- mutable values
            return self._copy(
                self._getter(),
            )

        else:
            # -- We use a deepcopy to ensure that we're not storing
            # -- mutable values
            return self._copy(
                getattr(
                    self._target(),
                    self._getter,
                ),
            )

    # --------------------------------------------------------------------------
    def set(self, *args, **kwargs):
        # -- If it s callable we call the method/function with the
        # -- value in the stored state as the argument
        if self._set_is_callable:
            self._setter(*args, **kwargs)

        # -- The setter is not callable, so we assume its a property
        # -- to be set directly
        else:
            setattr(
                self._target(),
                self.label,
                *args,
                **kwargs
            )



from ..serialiser import SerialiserBase

import os
import sys
import json


# ------------------------------------------------------------------------------
# -- OS System wrangling
_APPDATA_ROOT = None
if sys.platform == "win32":
    _APPDATA_ROOT = os.path.join(
        os.environ['APPDATA'],
        'recollection',
    )

elif sys.platform == "linux":
    _APPDATA_ROOT = os.path.join(
        os.path.expanduser('~'),
        'AppData',
        'recollection',
    )

else:
    raise NotImplementedError(
        'Support for %s is not yet implemented' % sys.platform
    )


# ------------------------------------------------------------------------------
class JsonAppSerialiser(SerialiserBase):
    """
    This will serialise the recollection object to a json file and store it in
    an application data location (varies depending on platform). You can
    utilise the identifier as a sub-path to create namespaced serialisation
    objects.

    For instance:
        serialiser.serialise(recollection, 'foo/bar'

    would (on windows) result in:
        %APPDATA%/recollection/foo/bar.json

    """

    # --------------------------------------------------------------------------
    @staticmethod
    def serialise(data, identifier):
        path = JsonAppSerialiser.locator(identifier)

        directory = os.path.dirname(path)

        if not os.path.exists(directory):
            os.makedirs(directory)

        with open(path, 'w') as f:
            json.dump(
                data,
                f,
                indent=4,
                sort_keys=True,
            )

    # --------------------------------------------------------------------------
    @staticmethod
    def deserialise(identifier):
        path = JsonAppSerialiser.locator(identifier)

        if not os.path.exists(path):
            return None

        with open(path, 'r') as f:
            return json.load(f)

    # --------------------------------------------------------------------------
    @staticmethod
    def locator(identifier):
        """
        With the given identifier this should return the resolved location
        of the data

        :param identifier: This is very dependent on the requirements of
            each individual serialiser. This could be a filepath, a url,
            a name etc.
        :type identifier: str

        :return: str
        """
        parts = identifier.split('/')
        parts[-1] = parts[-1] + '.json'
        path = _APPDATA_ROOT

        while parts:
            path = os.path.join(
                path,
                parts.pop(0),
            )

        return path



"""
This demo attempts to show how we can user Inference within Memento 
to make setting up states trivial. It also shows how we might have 
an object with lots of inter-twining properties, and therefore is a
good example of where we can reliably roll-back at an object level.

It also shows using a mix of getters and setters in a more pracical
example.
"""
import math
import recollection


# ------------------------------------------------------------------------------
def demo():

    # -- Instance a pin
    pin = Pin()

    # -- Move the pin multiple times
    pin.move_to([1, 1], 1)
    pin.move_to([2, 2], 2)

    # -- Print our interim details so we can compare later
    print('Interim Destination : %s' % pin.position())
    print('\tDistance Moved : %s' % pin.distance_moved)
    print('\tTime Moving : %s' % pin.time_travelled)

    # -- Do a final move
    pin.move_to([10, 10], 1)

    # -- Print the current position and distance moved
    print('\nEnd Destination : %s' % pin.position())
    print('\tDistance Moved : %s' % pin.distance_moved)
    print('\tTime Moving : %s' % pin.time_travelled)

    # -- Now lets do one roll-back
    pin.memento.restore(1)

    # -- Print the fact that we have done one restore
    # -- step but all values are correctly rolled back
    # -- together
    print('\nRolled Back Destination : %s' % pin.position())
    print('\tDistance Moved : %s' % pin.distance_moved)
    print('\tTime Moving : %s' % pin.time_travelled)


# ------------------------------------------------------------------------------
class Pin(recollection.Inference):

    def __init__(self):
        super(Pin, self).__init__()

        # -- Public Properties
        self.distance_moved = 0
        self.time_travelled = 0

        # -- Private/Complex Properies
        self._current_position = [0, 0]

        self.memento.register(
            [
                'distance_moved',
                'time_travelled',
            ]
        )

    # --------------------------------------------------------------------------
    @recollection.infer.get('position')
    def position(self):
        return self._current_position

    # --------------------------------------------------------------------------
    @recollection.infer.store('position')
    def move_to(self, position, speed=1):

        # -- Because this method is already decorated, it will store
        # -- state on exit. Therefore we want to tell memento not
        # -- to store any per-property changes which means we get all
        # -- the changes in a single step
        with self.memento.omit():

            # -- Calculate the distance travelled
            travel_distance = self._distance(
                self._current_position,
                position,
            )

            # -- Set the distance travelled, calcualte the time
            # -- travelled based on the speed and update the position.
            # -- This is a good example of multiple properties being
            # -- closely coupled which are important to keep in step
            # -- with one-another should we want to 'undo' our move.
            self.distance_moved += travel_distance
            self.time_travelled += travel_distance / float(speed)
            self._current_position = position

    # --------------------------------------------------------------------------
    @classmethod
    def _distance(cls, p1, p2):
        distance = math.sqrt(
            math.pow(p2[0] - p1[0], 2) + math.pow(p2[1] - p1[1], 2),
        )

        return distance


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    demo()



"""
This demonstrates how we can use Memento to store the state of players
'pieces' on a board game. To demonstrate state history we assume on player
is an adult and one player a child - and therefore we want the child to 
be allowed to 'undo' their turn if they are not happy!
"""
import random
import recollection


# ------------------------------------------------------------------------------
def demo():
    """
    This will run a little demo example which prints out information about 
    each turn of a randomised game. 
    
    The dice rolls are random - as are the consequences of the turn. However
    the result will always be the 'child' winning the game due to the game
    being rolled-back (using memento) whenever the child is unhappy with their
    dice roll! 
    
    :return: 
    """
    # -- Start by instancing a game
    game = GameBoard('demo_game')

    # -- Add two players
    parent = game.add_player('Parent')
    child = game.add_player('Child')

    # -- Define this variable simply for readability in later
    # -- logic tests.
    good_outcome = True

    # -- Track our turn count so we can log it as we go
    turn_count = 1

    # -- Start the turn cycle!
    while not game.has_a_winner():

        print('starting turn %s...' % turn_count)

        for player in game.players.values():

            # - Roll the dice for the parent, and inform the game board
            outcome = game.move_player(
                player.name,
                player.roll_the_dice(),
            )

            # -- If the child is unhappy with their turn we must
            # -- revert the game state to what it was before the
            # -- dice roll!
            if player == child and outcome != good_outcome:
                game.state.restore(index=1)
                print('%s is undoing their turn!' % player.name)

        # -- Increment the turn count
        turn_count += 1


# ------------------------------------------------------------------------------
class GameBoard(object):
    """
    This is our gameboard object which tracks player progress throughout
    their game. We track position on the board and lives. Our win state
    for this demo is simply whoever survives!
    """
    def __init__(self, name_of_game):

        # - Store the name of the game
        self.name_of_game = name_of_game

        # -- We will use these variables to track
        # -- player progress
        self.players = dict()
        self.player_positions = dict()
        self.player_lives = dict()

        # -- Define our memento stack, allowing us to store state
        # -- and retrieve history
        self.state = recollection.Memento(self)

        # -- Define what we want to store from our game board
        self.state.register('player_lives')
        self.state.register('player_positions')
        self.state.register('name_of_game')

        # -- Define a serialiser - so that we can put the game
        # -- away and come back to it later if we want!
        self.state.register_serialiser(
            recollection.JsonAppSerialiser,
            'examples/boardgame/%s' % name_of_game
        )

        # -- Store the state as the zero state
        self.state.store()

    # --------------------------------------------------------------------------
    def add_player(self, name):
        """
        Adds a new player to the game, intialises the internal tracking
        variables etc
        """
        self.players[name] = Player(name)
        self.player_positions[name] = 0
        self.player_lives[name] = 5

        return self.players[name]

    # --------------------------------------------------------------------------
    def move_player(self, player_name, squares_to_move):
        """
        Moves the player by the given number, then prints a consequence
        and returns True if the consquence was positive and false if the
        consquence was negative.

        Crucially we store the state on ever roll - so if we need
        to go back, we can do.
        """
        self.player_positions[player_name] += squares_to_move

        # -- Randomly pick a consquence!
        result = random.randrange(0, 4)
        outcome = True

        if result < 2:
            # -- No consequnce
            pass

        elif result == 3:
            print('\t%s has been bitten by a snake!' % player_name)
            self.player_lives[player_name] -= 1
            outcome = False

        elif result == 4:
            print('\t%s has fallen down a rabbit hole!' % player_name)
            self.player_lives[player_name] -= 1
            outcome = False

        # -- Store the state after the move
        self.state.store()

        return outcome

    # --------------------------------------------------------------------------
    def has_a_winner(self):
        """
        Checks whether any players are dead and should be removed as well as 
        checking for a win state.
        """
        player_names = [name for name in self.players.keys()]

        for player_name in player_names:

            if self.player_lives[player_name] <= 0:
                self.players.pop(player_name)
                self.player_lives.pop(player_name)
                self.player_positions.pop(player_name)
                print('\t\t%s is out the game!' % player_name)

            else:
                print(
                    '\t\t%s has %s lives left!' % (
                        player_name,
                        self.player_lives[player_name],
                    )
                )

        if len(self.players) == 1:
            print('%s has won!' % list(self.players.values())[0].name)
            return True

        return False


# ------------------------------------------------------------------------------
class Player(object):
    """
    This class represents a player
    """

    # --------------------------------------------------------------------------
    def __init__(self, name):
        self.name = name

    # --------------------------------------------------------------------------
    def roll_the_dice(self):
        roll = random.randrange(1, 6)

        print(
            '\t%s has rolled %s' % (
                self.name,
                roll,
            )
        )

        return roll


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    demo()



import abc
import six


# ------------------------------------------------------------------------------
@six.add_metaclass(abc.ABCMeta)
class SerialiserBase(object):
    """
    A serialiser is responsible for serialising the state of a momento
    object in such a way that it can easily be restored from the state
    of persistence.

    Note: Different serialisers have different levels of support for
    different datatypes. Therefore please read the docstrings of the
    serialisers when deciding which serialiser you want to utilise.
    """

    # --------------------------------------------------------------------------
    @staticmethod
    @abc.abstractmethod
    def serialise(data, identifier):
        """
        Attempts to serialise the given memento using the identifier
        variable to determine where the serialisation artifact should
        be stored.

        :param data: This is a dictionary snapshot from a memento stack
        :type data: dict

        :param identifier: This is very dependent on the requirements of
            each individual serialiser. This could be a filepath, a url,
            a name etc.
        :type identifier: str

        :return: True if the serialisation was successful
        """
        pass

    # --------------------------------------------------------------------------
    @staticmethod
    @abc.abstractmethod
    def deserialise(identifier):
        """
        Sets the state of the memento object based on the serialised data

        :param identifier: This is very dependent on the requirements of
            each individual serialiser. This could be a filepath, a url,
            a name etc.
        :type identifier: str

        :return: Dictionary which can be inserted into a memento stack
        """
        pass

    # --------------------------------------------------------------------------
    @staticmethod
    @abc.abstractmethod
    def locator(identifier):
        """
        With the given identifier this should return the resolved location
        of the data

        :param identifier: This is very dependent on the requirements of
            each individual serialiser. This could be a filepath, a url,
            a name etc.
        :type identifier: str

        :return: str
        """
        pass



from recollection.tests.classes import (
    EmptyTestClass,
    SetterTestClass,
    EventExceptingClass,
    EventCalledException,
)

import recollection
import unittest


# ------------------------------------------------------------------------------
class TestMemento(unittest.TestCase):
    """
    This suite of tests aims to give full coverage over memento.core by
    testing all the functionality within the object.
    """

    # --------------------------------------------------------------------------
    def test_can_instance_memento(self):
        """
        Ensures the memento object can be instanced
        :return:
        """
        test_class = EmptyTestClass()
        recollection.Memento(test_class)

    # --------------------------------------------------------------------------
    def test_can_store(self):
        """
        Ensures that we can store data, and that data is held in
        the memento object

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register('foo')

        # -- Set the value and store the result
        test_class.foo = 2
        stack.store()

        # -- Check that we stored 2
        self.assertEqual(
            2,
            stack._states[0]['foo'],
        )

    # --------------------------------------------------------------------------
    def test_can_restore(self):
        """
        Ensures that we can restore data from a memento state

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register('foo')

        # -- Store the current state
        stack.store()

        # -- Set the value and store the result
        test_class.foo = 2

        # -- Restore the last state
        stack.restore()

        # -- Check that we stored 2
        self.assertEqual(
            1,
            test_class.foo,
        )

    # --------------------------------------------------------------------------
    def test_can_restore_historically(self):
        """
        Ensures we can rollback to an older version of data

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register('foo')

        for i in range(11):
            test_class.foo = i

            # -- Store the current state
            stack.store()

        # -- Test that the latest value is indeed 10
        self.assertEqual(
            10,
            test_class.foo
        )

        # -- Restore 5 versions back
        stack.restore(index=5)

        self.assertEqual(
            5,
            test_class.foo,
        )

    # --------------------------------------------------------------------------
    def test_lock_step(self):
        """
        Ensures that the lock-step mechanism is working correctly

        :return:
        """
        # -- Create a class and define a property
        test_class_a = EmptyTestClass()
        test_class_b = EmptyTestClass()
        test_class_a.foo = 1
        test_class_b.foo = 1

        # -- Create the memento object
        stack_a = recollection.Memento(test_class_a)
        stack_b = recollection.Memento(test_class_b)

        stack_a.register('foo')
        stack_b.register('foo')

        # -- Group the stacks
        stack_a.group(stack_b)

        for i in range(11):
            test_class_a.foo = i
            test_class_b.foo = i

            # -- Store once
            stack_a.store()

        # -- Rollback
        stack_a.restore(5)

        self.assertEqual(
            5,
            test_class_a.foo,
        )

        self.assertEqual(
            5,
            test_class_b.foo,
        )

    # --------------------------------------------------------------------------
    def test_adding_to_group_multiple_times(self):
        """
        Groups allow lock-stepping. This test ensures all the logic which
        ensures instances do not exist multiple times in a group

        :return:
        """
        # -- Create a class and define a property
        test_class_a = EmptyTestClass()
        test_class_b = EmptyTestClass()
        test_class_c = EmptyTestClass()

        # -- Create the memento object
        stack_a = recollection.Memento(test_class_a)
        stack_b = recollection.Memento(test_class_b)
        stack_c = recollection.Memento(test_class_c)

        # -- Group the stacks
        stack_a.group(stack_b)
        stack_a.group(stack_b)
        stack_a.group(stack_b)
        stack_a.group(stack_b)

        stack_b.group(stack_a)
        stack_b.group(stack_a)
        stack_b.group(stack_a)
        stack_b.group(stack_a)

        # -- Side test to check when group a is in a group already
        # -- and group c is not
        stack_a.group(stack_c)

        self.assertEqual(
            3,
            len(stack_a.sync_group())
        )

    # --------------------------------------------------------------------------
    def test_can_unregister(self):
        """
        Checks that when we unregister something, we no longer look for
        it or utilise it

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register('foo')

        # -- Set the value and store the result
        test_class.foo = 2
        stack.store()

        # -- Check that we stored 2
        self.assertEqual(
            2,
            stack._states[0]['foo'],
        )

        # -- Now unregister foo
        stack.unregister('foo')
        stack.store()

        self.assertNotIn(
            'foo',
            stack._states[0],
        )

    # --------------------------------------------------------------------------
    def test_setter_methods(self):
        """
        Checks to see that the explicit declaration of setters is working

        :return:
        """
        # -- Create our test class
        test_class = SetterTestClass()

        # -- Instance a stack
        stack = recollection.Memento(test_class)

        # -- Register an attribute using getters and setters
        stack.register(
            label='target',
            getter=test_class.getTarget,
            setter=test_class.setTarget,
        )

        # -- Set the target
        test_class.setTarget(42)
        stack.store()

        # -- Set the target again
        self.assertEqual(
            42,
            stack._states[0]['target'],
        )

    # --------------------------------------------------------------------------
    def test_object_copying(self):
        """
        When we store state we have the option of specifically copying objects
        or just referencing them.

        In most cases we want to copy, as state is dependent on it - but not
        always. Therefore this is optional and this tests specifically that
        we get new objects when copying.

        :return:
        """
        # -- Create an object
        generic_object = EmptyTestClass()

        # -- Create our test class
        test_class = SetterTestClass()

        # -- Instance the memento stack
        stack = recollection.Memento(test_class)

        # -- Register an object to store
        stack.register(
            label='target',
            getter=test_class.getTarget,
            setter=test_class.setTarget,
            copy_value=True
        )

        # -- Set the attribute to our object - which should
        # -- create a copy each time
        for i in range(11):
            test_class.setTarget(generic_object)
            stack.store()

        # -- Fail if they are the same
        self.assertNotEqual(
            generic_object,
            stack._states[0]['target'],
        )

    # --------------------------------------------------------------------------
    def test_object_not_copying(self):
        """
        When we store state we have the option of specifically copying objects
        or just referencing them.

        In most cases we want to copy, as state is dependent on it - but not
        always. Therefore this is optional and this tests specifically that
        we get the same objects when not copying.

        :return:
        """
        # -- Create the object which we will ultimately test against
        generic_object = EmptyTestClass()

        # -- Create our test class
        test_class = SetterTestClass()

        # -- Instance a memento stack
        stack = recollection.Memento(test_class)

        # -- Regiser the attribute
        stack.register(
            label='target',
            getter=test_class.getTarget,
            setter=test_class.setTarget,
            copy_value=False
        )

        # -- Call the store a few times - this time we're expecting
        # -- the object to be referenced
        for i in range(11):
            test_class.setTarget(generic_object)
            stack.store()

        # -- Fail if the objects are not the same
        self.assertEqual(
            generic_object,
            stack._states[0]['target'],
        )

    # --------------------------------------------------------------------------
    def test_simple_list_registration(self):
        """
        Memento allows us to specify a list of attributes by string name, this
        test ensures that mechanism works.

        :return:
        """
        # -- Define our list of properties
        properties = [
            'a',
            'b',
            'c',
        ]

        # -- Create a class and define a property
        test_class = EmptyTestClass()

        # -- Set values on each property
        for prop in properties:
            setattr(
                test_class,
                prop,
                1,
            )

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register(properties)
        stack.store()

        # -- Ensure that each property is in the stack
        for prop in properties:
            self.assertIn(
                prop,
                stack._states[0],
            )

    # --------------------------------------------------------------------------
    def test_complex_list_registration(self):
        """
        As well as giving a list of strings we can also give a list containing
        lists of index-ordered parameters. This ensures that mechanism works
        as expected.

        :return:
        """
        # -- Create a class and define a property
        test_class = SetterTestClass()

        # -- Create the memento object
        stack = recollection.Memento(test_class)

        # -- Register the two getters and setters in one go
        stack.register(
            [
                [
                    test_class.getTarget,
                    test_class.setTarget,
                    'target',
                    True,
                ],
                [
                    test_class.getDistance,
                    test_class.setDistance,
                    'distance',
                    True,
                ],
            ]
        )
        stack.store()

        # -- Ensure the target was stored
        self.assertIn(
            'target',
            stack._states[0],
        )

        # -- Ensure the distance was stored
        self.assertIn(
            'distance',
            stack._states[0],
        )

    # --------------------------------------------------------------------------
    def test_assert_on_label_duplication(self):
        """
        This is a negative test to ensure we get an assertion when we
        try to register the same label twice.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)

        # -- Register a label
        stack.register('foo')

        # -- Try to register the same label again
        try:
            stack.register('foo')

            self.assertTrue(
                False,
                msg='Did not fail on registering the same label twice',
            )

        except recollection.RegistrationError:
            pass

    # --------------------------------------------------------------------------
    def test_out_of_scope_assert(self):
        """
        This will test teh assertion when our target object we're looking
        at is garbage collected.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)
        stack.register('foo')

        # -- Delete the object
        # noinspection PyRedundantParentheses
        del(test_class)

        # -- Attempt to store using the stack
        try:
            stack.store()

            self.assertTrue(
                False,
                msg='Did not fail on storing a deleted object reference',
            )

        except recollection.StorageError:
            pass

    # --------------------------------------------------------------------------
    def test_type_error_on_invalid_label(self):
        """
        This ensures that we get a type error if giving an invalid
        label type.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class)

        try:
            # noinspection PyTypeChecker
            stack.register(
                label=123,
                getter='foo',
            )
            self.assertTrue(
                False,
                msg='Did not fail when giving an invalid label type',
            )

        except TypeError:
            pass

    # --------------------------------------------------------------------------
    def test_max_states_are_popped(self):
        """
        We can define the max states to store on a stack. This checks that
        the behaviour when exceeding that value is to have the stack constantly
        popping.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)
        stack.register('foo')

        for i in range(100):
            test_class.foo = i
            stack.store()

        self.assertEqual(
            stack.count(),
            5,
        )

    # --------------------------------------------------------------------------
    def test_assert_registering_invalid_serialiser(self):
        """
        This will ensure that we get teh correct assert when giving a
        serialiser which does not base off our SerialiserBase

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()
        test_class.foo = 1

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)
        stack.register('foo')

        try:
            stack.register_serialiser(
                serialiser=EmptyTestClass(),
                identifier='',
            )

            self.assertTrue(
                False,
                msg='Did not fail when registering invalid serialiser',
            )

        except TypeError:
            pass

    # --------------------------------------------------------------------------
    def test_serialising_with_no_serialiser(self):
        """
        Tests that we get an assert when trying to serialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)

        # noinspection PyBroadException
        try:
            stack.serialise()

            self.assertTrue(
                False,
                msg='Did not fail when serialising with no serialiser',
            )

        except Exception:
            pass

    # --------------------------------------------------------------------------
    def test_deserialising_with_no_serialiser(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EmptyTestClass()

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)

        # noinspection PyBroadException
        try:
            stack.deserialise()

            self.assertTrue(
                False,
                msg='Did not fail when serialising with no serialiser',
            )

        except Exception:
            pass

    # --------------------------------------------------------------------------
    def test_registration_event_triggering(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EventExceptingClass()
        test_class.foo = 0

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)

        # -- Hook up the event
        stack.registered.connect(test_class.raise_exception)

        # noinspection PyBroadException
        try:
            stack.register('foo')

            self.assertTrue(
                False,
                msg='Event not triggered during exception',
            )

        except EventCalledException:
            pass

    # --------------------------------------------------------------------------
    def test_unregister_event_triggering(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EventExceptingClass()
        test_class.foo = 0

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)

        # -- Hook up the event
        stack.unregistered.connect(test_class.raise_exception)
        stack.register('foo')

        # noinspection PyBroadException
        try:
            stack.unregister('foo')

            self.assertTrue(
                False,
                msg='Event not triggered during exception',
            )

        except EventCalledException:
            pass

    # --------------------------------------------------------------------------
    def test_store_event_triggering(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EventExceptingClass()
        test_class.foo = 0

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)
        stack.register('foo')

        # -- Hook up the event
        stack.stored.connect(test_class.raise_exception)

        # noinspection PyBroadException
        try:
            stack.store()

            self.assertTrue(
                False,
                msg='Event not triggered during exception',
            )

        except EventCalledException:
            pass

    # --------------------------------------------------------------------------
    def test_restore_event_triggering(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EventExceptingClass()
        test_class.foo = 0

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)
        stack.register('foo')
        stack.store()
        stack.store()

        # -- Hook up the event
        stack.restored.connect(test_class.raise_exception)

        # noinspection PyBroadException
        try:
            stack.restore()

            self.assertTrue(
                False,
                msg='Event not triggered during exception',
            )

        except EventCalledException:
            pass

    # --------------------------------------------------------------------------
    def test_unregister_event_triggering(self):
        """
        Tests that we get an assert when trying to deserialise with no
        serialiser registered.

        :return:
        """
        # -- Create a class and define a property
        test_class = EventExceptingClass()
        test_class.foo = 0

        # -- Create the memento object
        stack = recollection.Memento(test_class, max_states=5)

        # -- Hook up the event
        stack.registered.connect(test_class.raise_exception)
        stack.registered.disconnect(test_class.raise_exception)

        # noinspection PyBroadException
        try:
            stack.register('foo')

        except EventCalledException:
            self.assertTrue(
                False,
                msg='Called event after it was disconnected',
            )



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='recollection',
    version='0.9.3',
    author='Mike Malinowski',
    author_email='mike@twisted.space',
    description='A python package exposing the memento design pattern',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mikemalinowski/recollection',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'six',
    ],
    keywords="recollect recollection memento mementos",
)
